__all__ = [
    'decompressor',
    'compressor',
    'signatures',
    'decompress',
    'compress'
]

from sc_compression.compressor import Compressor
from sc_compression.decompressor import Decompressor


def compress(buffer: bytes, signature: int, file_version: int = None) -> bytes:
    return Compressor().compress(buffer, signature, file_version)


def decompress(buffer: bytes) -> (bytes, int):
    decompressor = Decompressor()
    decompressed = decompressor.decompress(buffer)

    return decompressed, decompressor.signatures.last_signature



import re


class Signatures:
    LZMA = 2 ** 0
    SC = 2 ** 1
    SCLZ = 2 ** 2
    SIG = 2 ** 3
    ZSTD = 2 ** 4
    NONE = 2 ** 5

    def __init__(self):
        self.last_signature: int = -1

    def get_signature(self, buffer, file_version: int = -1) -> int:
        signature = Signatures.NONE
        if self.last_signature == Signatures.SC and file_version == 1:
            signature = Signatures.LZMA
            self.last_signature = -1
        elif buffer[:4] == b'\x28\xb5\x2f\xfd':  # self.last_signature == Signatures.SC and file_version >= 2:
            signature = Signatures.ZSTD
            self.last_signature = -1

        if re.match(b'\x00\x00?\x00', buffer[1:5]):
            signature = Signatures.LZMA
        if buffer.startswith(b'SCLZ'):
            signature = Signatures.SCLZ
        elif buffer.startswith(b'SC'):
            signature = Signatures.SC
        elif buffer[:4] == b'Sig:':
            signature = Signatures.SIG

        self.last_signature = signature

        return signature



from sc_compression.utils.writer import Writer
from tempfile import mktemp
from os import remove, system, path


class LZHAM:
    @staticmethod
    def decompress(data, uncompressed_size, filters):
        writer = Writer('little')
        writer.write(b'LZH\x30')
        writer.writeByte(filters['dict_size_log2'])
        writer.writeUInt32(uncompressed_size)
        writer.writeUInt32(0)
        writer.write(data)

        temp_file_path = mktemp('.lzham')
        with open(temp_file_path, 'wb') as f:
            f.write(writer.buffer)
            f.close()

        if system(f'{path.dirname(__file__)}/lzham.exe '
                  f'-c -d{filters["dict_size_log2"]} '
                  f'd {temp_file_path} {temp_file_path} > nul 2>&1'):
            return None
        with open(temp_file_path, 'rb') as f:
            decompressed = f.read()
            f.close()

        remove(temp_file_path)

        return decompressed

    @staticmethod
    def compress(data, filters):
        temp_file_path = mktemp('.data')
        with open(temp_file_path, 'wb') as f:
            f.write(data)
            f.close()

        compressed_path = mktemp('.lzham')
        if system(f'{path.dirname(__file__)}/lzham.exe '
                  f'-c -d{filters["dict_size_log2"]} '
                  f'c {temp_file_path} {compressed_path} > nul 2>&1'):
            return None
        with open(compressed_path, 'rb') as f:
            compressed = f.read()[13:]
            f.close()

        remove(temp_file_path)
        remove(compressed_path)

        return compressed



import lzma

from sc_compression.signatures import Signatures
from sc_compression.utils.reader import Reader

try:
    import lzham
except ImportError:
    from platform import system as get_system_name

    lzham = None
    if get_system_name() == 'Windows':
        from sc_compression.support.lzham import LZHAM

        lzham = LZHAM

try:
    import zstandard
except ImportError:
    zstandard = None


class Decompressor(Reader):
    def __init__(self):
        super().__init__(b'')
        self.signatures = Signatures()
        self.file_version = -1
        self.hash = None

    def decompress(self, buffer: bytes) -> bytes:
        super().__init__(buffer, 'little')

        decompressed = buffer
        signature = self.signatures.get_signature(self.buffer, self.file_version)
        if signature == Signatures.NONE:
            return decompressed

        if signature == Signatures.SC:
            super().__init__(buffer, 'big')
            self.read(2)
            self.file_version = self.readInt32()
            self.hash = self.read(self.readInt32())
            decompressed = self.decompress(buffer[26:])
        elif signature == Signatures.SIG:
            buffer = buffer[68:]
            decompressed = self.decompress(buffer)
        elif signature == Signatures.SCLZ:
            self.read(4)
            dict_size_log2 = self.readUByte()
            uncompressed_size = self.readInt32()

            if lzham:
                filters = {
                    'dict_size_log2': dict_size_log2
                }
                decompressed = lzham.decompress(self.buffer[self.tell():], uncompressed_size, filters)
        elif signature == Signatures.LZMA:
            decompressor = lzma.LZMADecompressor()
            compressed = self.buffer[:5] + b'\xff' * 8 + self.buffer[9:]

            decompressed = decompressor.decompress(compressed)
        elif signature == Signatures.ZSTD:
            if zstandard:
                decompressor = zstandard.ZstdDecompressor()
                decompressed = decompressor.decompress(self.buffer)
        else:
            raise TypeError(signature)

        return decompressed



import lzma
from hashlib import md5

from sc_compression.signatures import Signatures
from sc_compression.utils.writer import Writer

try:
    import lzham
except ImportError:
    from platform import system as get_system_name

    lzham = None
    if get_system_name() == 'Windows':
        from sc_compression.support.lzham import LZHAM

        lzham = LZHAM

try:
    import zstandard
except ImportError:
    zstandard = None


class Compressor(Writer):
    lzham_filters = {
        'dict_size_log2': 18
    }
    lzma_filters = [
        {
            "id": lzma.FILTER_LZMA1,
            "dict_size": 256 * 1024,
            "lc": 3,
            "lp": 0,
            "pb": 2,
            "mode": lzma.MODE_NORMAL
        },
    ]

    def __init__(self):
        super().__init__('little')

    def compress(self, data, signature: int, file_version: int = None) -> bytes:
        uncompressed_size = len(data)

        if file_version is None:
            file_version = 2 if zstandard and not (Signatures.SCLZ & signature) else 1

        if Signatures.ZSTD & signature and not zstandard or \
           Signatures.SCLZ & signature and not lzham:
            signature = Signatures.SC

        super().__init__('little')
        if signature is Signatures.NONE:
            return data
        elif ((Signatures.LZMA | Signatures.SIG) & signature) or (Signatures.SC & signature and file_version != 2):
            compressed = lzma.compress(data, format=lzma.FORMAT_ALONE, filters=self.lzma_filters)

            self.write(compressed[:5])

            self.writeInt32(uncompressed_size)

            self.write(compressed[13:])

            compressed = self.buffer
        elif (Signatures.SCLZ & signature) and lzham:
            compressed = lzham.compress(data, filters=self.lzham_filters)

            self.write(b'SCLZ')
            self.writeUByte(self.lzham_filters['dict_size_log2'])
            self.writeInt32(uncompressed_size)
            self.write(compressed)

            compressed = self.buffer
        elif (Signatures.SC | Signatures.ZSTD) & signature and file_version == 2:
            compressor = zstandard.ZstdCompressor()
            compressed = compressor.compress(data)
        else:
            raise TypeError('Unknown Signature!')

        super().__init__('big')
        if (Signatures.SC | Signatures.SCLZ) & signature:
            data_hash = md5(data)

            self.write(b'SC')
            self.writeInt32(file_version)
            self.writeInt32(16)
            compressed = self.buffer + data_hash.digest() + compressed
        elif signature == Signatures.SIG:
            self.write(b'Sig:')
            self.write(b'\x00' * 64)  # sha64
            compressed = self.buffer + compressed

        return compressed
